## Brian Blaylock
## July 8, 2021

import os
import warnings
from pathlib import Path

__author__ = "Brian K. Blaylock"


try:
    ## TODO: Will the `_version.py` file *always* be present?
    ## TODO: What if the person doesn't do "pip install"
    from ._version import __version__, __version_tuple__
except:
    __version__ = "unknown"
    __version_tuple__ = (999, 999, 999)


import toml

# =======================================================================
# Load custom xarray accessors
# TODO: Move some of the tools.py to these accessors.
try:
    import goes2go.accessors
except:
    warnings.warn("goes2go xarray accessors could not be imported.")


# =======================================================================
# Overload Path object with my custom `expand` method so the user can
# set environment variables in the config file (e.g., ${HOME}).
def _expand(self):
    """
    Fully expand and resolve the Path with the given environment variables.

    Example
    -------
    >>> Path('$HOME').expand()
    PosixPath('/p/home/blaylock')
    """
    return Path(os.path.expandvars(self)).expanduser().resolve()


Path.expand = _expand

# =======================================================================
# Location of goes2go configuration file
_config_path = os.getenv("GOES2GO_CONFIG_PATH", "~/.config/goes2go")
_config_path = Path(_config_path).expand()
_config_file = _config_path / "config.toml"

# Default directory goes2go saves model output
# NOTE: The `\\` is an escape character in TOML.
#       For Windows paths, "C:\\user\\"" needs to be "C:\\\\user\\\\""
_save_dir = os.getenv("GOES2GO_SAVE_DIR", "~/data")
_save_dir = Path(_save_dir).expand()
_save_dir = str(_save_dir).replace("\\", "\\\\")

# =======================================================================
# Default TOML Configuration
default_toml = f""" # GOES-2-go Defaults
["default"]
save_dir = "{_save_dir}"
satellite = "noaa-goes16"
product = "ABI-L2-MCMIP"
domain = "C"
download = true
return_as = "filelist"
overwrite = false
max_cpus = 1
s3_refresh = true
verbose = true

["timerange"]
s3_refresh = false

["latest"]
return_as = "xarray"

["nearesttime"]
within = "1H"
return_as = "xarray"
"""

########################################################################
# Load config file (create one if needed)
try:
    # Load the goes2go config file
    config = toml.load(_config_file)
except:
    try:
        # Create the goes2go config file
        _config_path.mkdir(parents=True, exist_ok=True)
        with open(_config_file, "w", encoding="utf-8") as f:
            f.write(default_toml)

        print(
            f" ╭─goes2go──────────────────────────────────────────────╮\n"
            f" │ INFO: Created a default config file.                 │\n"
            f" │ You may view/edit goes2go's configuration here:      │\n"
            f" │ {str(_config_file):^53s}│\n"
            f" ╰──────────────────────────────────────────────────────╯\n"
        )

        # Load the new goes2go config file
        config = toml.load(_config_file)
    except (FileNotFoundError, PermissionError, IOError):
        print(
            f" ╭─goes2go─────────────────────────────────────────────╮\n"
            f" │ WARNING: Unable to create config file               │\n"
            f" │ {str(_config_file):^53s}│\n"
            f" │ goes2go will use standard default settings.         │\n"
            f" │ Consider setting env variable GOES2GO_CONFIG_PATH.  │\n"
            f" ╰─────────────────────────────────────────────────────╯\n"
        )
        config = toml.loads(default_toml)


# Expand the full path for `save_dir`
config["default"]["save_dir"] = Path(config["default"]["save_dir"]).expand()

if os.getenv("GOES2GO_SAVE_DIR"):
    config["default"]["save_dir"] = Path(os.getenv("GOES2GO_SAVE_DIR")).expand()
    print(
        f" ╭─goes2go──────────────────────────────────────────────╮\n"
        f" │ INFO: Overriding the configured save_dir because the │\n"
        f" │ environment variable GOES2GO_SAVE_DIR is set to      │\n"
        f" │ {os.getenv('GOES2GO_SAVE_DIR'):^53s}│\n"
        f" ╰──────────────────────────────────────────────────────╯\n"
    )

# Merge default settings with overwrite settings for each download method
for i in ["timerange", "latest", "nearesttime"]:
    config[i] = {**config["default"], **config[i]}


from goes2go.data import goes_latest, goes_nearesttime, goes_timerange
from goes2go.NEW import GOES



from goes2go import GOES


########################################################################


def test_GOES16():
    """Create a GOES object"""
    G = GOES(satellite="noaa-goes16", domain="C")
    assert G.satellite == "noaa-goes16"
    assert G.domain == "C"


def test_GOES16_latest():
    ds = GOES(satellite=16).latest()


def test_GOES16_nearesttime():
    ds = GOES(satellite=16).nearesttime("2022-01-01")


def test_GOES16_timerange():
    ds = GOES(satellite=16).timerange("2022-01-01 00:00", "2022-01-01 01:00")


def test_GOES16_df():
    df = GOES(satellite=16).df("2022-01-01 00:00", "2022-01-01 01:00")


########################################################################


def test_GOES18():
    """Create a GOES object"""
    G = GOES(satellite="noaa-goes18", domain="C")
    assert G.satellite == "noaa-goes18"
    assert G.domain == "C"


def test_GOES18_latest():
    ds = GOES(satellite=18).latest()


def test_GOES18_nearesttime():
    ds = GOES(satellite=18).nearesttime("2023-07-01")



## Brian Blaylock
## June 4, 2022

"""
==========
GOES Class
==========
"""

import itertools
import logging
import re
from pathlib import Path

import numpy as np
import pandas as pd
import s3fs
import toml

from goes2go import config
from goes2go.data import _goes_file_df, goes_latest, goes_nearesttime, goes_timerange

log = logging.getLogger(__name__)

tables_dir = Path(__file__).parent

# Connect to AWS public buckets
fs = s3fs.S3FileSystem(anon=True)

product_table = pd.read_csv(
    tables_dir / "product_table.txt",
    skiprows=2,
    names=["product", "description"],
    index_col=0,
)
product_table.index = product_table.index.str.strip()
product_table["description"] = product_table.description.str.strip()


# Assume goes17 and goes18 have same products as goes16
_product = {i.split("/")[-1] for i in fs.ls("noaa-goes16")}
_product = set(filter(lambda x: x.split(".")[-1] not in ["pdf", "html"], _product))
_product

# you can be unspecific and request any mesoscale domain (M),
# or by number (M1, M2)
_domains = {"F", "C", "M", "M1", "M2"}


class GOES:
    """The GOES satellite class."""

    def __init__(
        self,
        satellite=config["timerange"].get("satellite"),
        product=config["timerange"].get("product"),
        domain=config["timerange"].get("domain"),
        bands=None,
        channel=None,
    ):
        """Initialize a GOES object for a desired satellite and product.

        Parameters
        ----------
        satellite : {16, 17, 18}
            The satellite number. May also use the following aliases
            {'G16', "G17", "EAST", "WEST"}
        product : str
            The product to acquire. A full list of products is here
            https://github.com/blaylockbk/goes2go/blob/main/goes2go/product_table.txt
            - GLM = alias for geostationary lighting mapper
            - ABI = alias for ABI multi-channel cloud moisture imagery
        domain : {None, 'F', 'C', "M", "M1", "M2"}
            Only needed for ABI products.
            - F = Full Disk
            - C = CONUS
            - M = Mesoscale sector (both)
            - M1 = Mesoscale sector 1
            - M2 = Mesoscale sector 2
        bands : None, int, or list
            Specify the ABI channels to retrieve. *Only used if the
            product requested has unique bands.*
            For example, ``bands=2`` for channel 2 or ``bands=[1,2,3]``
            for channels 1, 2 and 3.
        channel : None, int, or list
            Alias for "bands" argument. If channel is not None, then
            "bands" is not use.

            Note: I don't like the name of "bands", but the NetCDF file
            uses the term "band" instead of "channel" in reference to
            the ABI products, so I'll stick with "band" for now.
        """
        self.satellite = satellite
        self.product = product
        self.domain = domain

        if channel is not None:
            self.bands = channel
        else:
            self.bands = bands

        if self.product.startswith("ABI") and self.product in _product:
            # Sometimes the user might inavertantly give the domain
            # in the product name.
            self.domain = self.product[-1]
            self.product = self.product[:-1]

        self._check_satellite()
        self._check_product()

    def _check_satellite(self):
        if isinstance(self.satellite, int):
            self.satellite = f"noaa-goes{self.satellite}"
        elif isinstance(self.satellite, str):
            if self.satellite.upper() == "EAST":
                self.satellite = "noaa-goes16"
            elif self.satellite.upper() == "WEST":
                self.satellite = "noaa-goes17"
            else:
                # look for the satellite number in the string (i.e.g, 'G16', 'goes16')
                self.satellite = re.sub("[^0-9]", "", self.satellite)
                self.satellite = f"noaa-goes{self.satellite}"
        else:
            raise ValueError(
                f"Could not figure out what satellite you want from `{self.satellite}`"
            )

    def _check_product(self):
        if self.product == "GLM":
            # Alias for geostationary lighting mapper
            self.product = "GLM-L2-LCFA"
        elif self.product == "ABI":
            # Alias for multi-channel cloud moisture imagery
            if self.domain is None:
                self.product = "ABI-L2-MCMIP" + "C"
            else:
                self.product = "ABI-L2-MCMIP" + re.sub("[0-9]", "", self.domain)
        elif self.product.startswith("ABI"):
            if self.domain is None:
                if self.product[-1] in _domains:
                    self.domain = self.product[-1]

            elif self.domain is not None:
                if self.domain in _domains:
                    self.product = self.product + re.sub("[0-9]", "", self.domain)
                    if self.product not in _product:
                        raise ValueError(f"{self.product} is not a valid product.")
                else:
                    raise ValueError(
                        f"domain for ABI products must be None or one of {_domains}"
                    )
        else:
            if self.domain is not None:
                log.warning("domain argument is ignored for non-ABI products")

        if self.product in _product:
            self.description = product_table.loc[self.product].description
        else:
            raise ValueError(f"{self.product} is not an available product.")

    def __repr__(self):
        msg = [
            "╭───────────────────────────────╌┄┈",
            "│ 🌎 GOES Object   ",
            "│ ───────────────",
            f"│  {self.satellite=}",
            f"│  {self.product=}",
            f"│  {self.domain=}",
            f"│  {self.bands=}",
            f"│  {self.description=}",
            "╰───────────────────────────────╌┄┈",
        ]
        return "\n".join(msg)

    def latest(self, **kwargs):
        """Get the latest available GOES data."""
        return goes_latest(
            satellite=self.satellite,
            product=self.product,
            domain=self.domain,
            bands=self.bands,
            **kwargs,
        )

    def nearesttime(
        self,
        attime,
        within=pd.to_timedelta(config["nearesttime"].get("within", "1H")),
        **kwargs,
    ):
        """Get the GOES data nearest a specified time.

        Parameters
        ----------
        attime : datetime
            Time to find the nearest observation for.
            May also use a pandas-interpretable datetime string.
        within : timedelta or pandas-parsable timedelta str
            Timerange tht the nearest observation must be.
        """
        return goes_nearesttime(
            attime,
            within,
            satellite=self.satellite,
            product=self.product,
            domain=self.domain,
            bands=self.bands,
            **kwargs,
        )

    def timerange(self, start=None, end=None, recent=None, **kwargs):
        """Get GOES data for a time range.

        Parameters
        ----------
        start, end : datetime
            Required if recent is None.
        recent : timedelta or pandas-parsable timedelta str
            Required if start and end are None. If timedelta(hours=1), will
            get the most recent files for the past hour.
        """
        return goes_timerange(
            start,
            end,
            recent,
            satellite=self.satellite,
            product=self.product,
            domain=self.domain,
            bands=self.bands,
            **kwargs,
        )

    def df(self, start, end, refresh=True):
        """Get list of requested GOES files as pandas.DataFrame.

        Parameters
        ----------
        start : datetime
        end : datetime
        refresh : bool
            Refresh the s3fs.S3FileSystem object when files are listed.
            Default True will refresh and not use a cached list.
        """
        return _goes_file_df(
            self.satellite,
            self.product,
            start=start,
            end=end,
            bands=self.bands,
            refresh=refresh,
        )
